#!/usr/bin/env python3

from gnuradio import gr
import numpy as np

import logging
import time
import signal
import sys

import polarimetric_interferometry as pol


class flowgraph(pol.flowgraph):
    def __init__(self, obs_time, *args, **kwargs):
        if gr.enable_realtime_scheduling() != gr.RT_OK:
            logging.warning("Error: failed to enable real-time scheduling.")
        super().__init__(*args, **kwargs)
        self.first_record = True
        self.obs_time = obs_time

        def sig_handler(sig=None, frame=None):
            self.stop()
            self.wait()
            logging.shutdown()
            sys.exit(0)

        signal.signal(signal.SIGINT, sig_handler)
        signal.signal(signal.SIGTERM, sig_handler)

    def start_record(self, source, baseline):
        params = {'src_name' : source,
                  'antennas' : '-'.join(baseline),
                  }
        logging.info(f'Starting recording of {params["src_name"]} with {params["antennas"]}')
        if self.first_record:
            self.first_record = False
            self.make_sinks(**params)
            self.start()
        else:
            self.reload_sinks(**params)
            self.unlock()

    def stop_record(self):
        logging.info(f'Stopping recording')
        self.lock()

    def observe(self, source, baseline):
        self.start_record(source, baseline)
        time.sleep(self.obs_time)
        self.stop_record()
        return self.get_power()

    def get_power(self):
        return [self.get_power_chan(chan) + self.get_power_chan(chan+1)
                for chan in range(0,4,2)]

    def get_power_chan(self, chan):
        path = self.sink_name(chan, 'power')
        x = np.fromfile(path, 'float32')
        x = x.reshape((-1, self.nfft))
        # throw away first 0.5 seconds
        x = x[5:]
        freqsel = np.arange(self.nfft//4, 3*self.nfft//4)
        return np.average(x[:, freqsel])
